import struct
import io


def sizeof(expr:dict|list|str,endian=">"):
    if type(expr)==str:
        return struct.calcsize(endian+expr)
    if type(expr)==list:
        return sum([sizeof(i) for i in expr])
    if type(expr)==dict:
        return sum([sizeof(v) for k,v in expr.items()])
    raise TypeError(expr)


class _Base:
    def __init__(self,file:io.BytesIO,endian=">",encoding="utf-8") -> None:
        self.file=file
        self.endian=endian
        self.encoding=encoding

class Reader(_Base):  
    def load(self,expr:dict|list|str):
        if type(expr)==str:
            x=self.file.read(struct.calcsize(expr))
            y=struct.unpack(self.endian+expr,x)[0]
            if self.encoding and type(y)==bytes:
                y=y.decode(self.encoding).partition("\x00")[0]
            return y
        if type(expr)==list:
            return [self.load(i) for i in expr]
        if type(expr)==dict:
            return {k:self.load(v) for k,v in expr.items()}
        raise TypeError(expr)

class Writer(_Base):  
    def dump(self,expr:dict|list|str,data:dict|list|str):
        if type(expr)==str:
            if self.encoding and type(data)==str:
                data=data.encode(self.encoding)
            x=struct.pack(self.endian+expr,data)
            y=self.file.write(x)
            return y
        if type(expr)==list:
            return [self.dump(i,j) for i,j in zip(expr,data)]
        if type(expr)==dict:
            return {k:self.dump(expr[k],data[k]) for k in expr}
        raise TypeError(expr)

class File(Reader,Writer):
    pass
